from sqlalchemy import dialects

# load sqlalchemy modules proxied through the models object
# this is mostly for convenience
from sqlalchemy import (
    MetaData,
    INT as Int,
    Integer,
    Float,
    Column,
    UniqueConstraint,
    Index,
    ForeignKey,
    String,
    DateTime,
    TIMESTAMP as TimeStamp,
    PickleType,
    Time,
    Table,
    or_,
    and_,
    not_,
    select
    )

from sqlalchemy.types import (
    BLOB,
    BOOLEAN,
    BigInteger,
    Binary,
    Boolean,
    CHAR,
    CLOB,
    DATE,
    DATETIME,
    DECIMAL,
    Date,
    DateTime,
    Enum,
    FLOAT,
    Float,
    INT,
    INTEGER,
    Integer,
    Interval,
    LargeBinary,
    NCHAR,
    NVARCHAR,
    NUMERIC,
    Numeric,
    PickleType,
    SMALLINT,
    SmallInteger,
    String,
    TEXT,
    TIME,
    TIMESTAMP,
    Text,
    Time,
    TypeDecorator,
    Unicode,
    UnicodeText,
    VARCHAR,
    )


from sqlalchemy.orm import (
    column_property,
    mapper,
    scoped_session,
    sessionmaker,
    reconstructor,
    relationship,
    backref,
    contains_eager,
    eagerload,
    eagerload_all,
    joinedload,
    synonym,
    object_session
    )

from sqlalchemy.orm.attributes import (
    instance_state
    )

from sqlalchemy.orm.util import (
    has_identity
)

from sqlalchemy.orm.exc import (
    NoResultFound,
    MultipleResultsFound
    )

from sqlalchemy.exc import (
    IntegrityError
    )

from sqlalchemy.sql import (
    exists,
    )

from sqlalchemy.sql.expression import (
    asc,
    case,
    desc,
    literal,
    literal_column,
    null
    )

from sqlalchemy.ext.hybrid import (
                    hybrid_property,
                    hybrid_method
    )

from sqlalchemy import func

import json
import pickle
import zlib
from pybald.db import engine, dump_engine
from pybald.util import camel_to_underscore, pluralize

import project
# from project import mc

# from sqlalchemy.orm.interfaces import SessionExtension
from sqlalchemy.ext.mutable import Mutable

from sqlalchemy import __version__ as sa_ver

sa_maj_ver, sa_min_ver, sa_rev_ver = sa_ver.split(".")

if int(sa_min_ver) >= 7:
    from sqlalchemy.orm.attributes import flag_modified
else:
    from sqlalchemy.orm.attributes import instance_dict, NO_VALUE

    def flag_modified(instance, key):
        state, dict_ = instance_state(instance), instance_dict(instance)
        impl = state.manager[key].impl
        state.modified_event(dict_, impl, True, NO_VALUE)


session_args = {}

if project.green:
    from SAGreen import eventlet_greenthread_scope
    session_args['scopefunc'] = eventlet_greenthread_scope

session = scoped_session(sessionmaker(bind=engine, **session_args))

import sqlalchemy.ext.declarative
from sqlalchemy.ext.declarative import declarative_base
Base = declarative_base(bind=engine)


# # lifted from zzzeek's post on "Magic ORM"
# # this only works for SA 7, due to the
# # use of the event system.
# # Source: http://techspot.zzzeek.org/2011/05/17/magic-a-new-orm/
# # Copyright (I assume) Michael Bayer
# # =====================================
# from sqlalchemy.orm import (
#             class_mapper, mapper, relationship,
#             scoped_session, sessionmaker, configure_mappers
#         )
# from sqlalchemy.ext.declarative import declared_attr, declarative_base
# from sqlalchemy import event
# import re
#
# @event.listens_for(mapper, "mapper_configured")
# def _setup_deferred_properties(mapper, class_):
#     """Listen for finished mappers and apply DeferredProp
#     configurations."""
#
#     for key, value in class_.__dict__.items():
#         if isinstance(value, DeferredProp):
#             value._config(class_, key)
#
# class DeferredProp(object):
#     """A class attribute that generates a mapped attribute
#     after mappers are configured."""
#
#     def _setup_reverse(self, key, rel, target_cls):
#         """Setup bidirectional behavior between two relationships."""
#
#         reverse = self.kw.get('reverse')
#         if reverse:
#             reverse_attr = getattr(target_cls, reverse)
#             if not isinstance(reverse_attr, DeferredProp):
#                 reverse_attr.property._add_reverse_property(key)
#                 rel._add_reverse_property(reverse)
#
# class FKRelationship(DeferredProp):
#     """Generates a one to many or many to one relationship."""
#
#     def __init__(self, target, fk_col, **kw):
#         self.target = target
#         self.fk_col = fk_col
#         self.kw = kw
#
#     def _config(self, cls, key):
#         """Create a Column with ForeignKey as well as a relationship()."""
#
#         target_cls = cls._decl_class_registry[self.target]
#
#         pk_target, fk_target = self._get_pk_fk(cls, target_cls)
#         pk_table = pk_target.__table__
#         pk_col = list(pk_table.primary_key)[0]
#
#         if hasattr(fk_target, self.fk_col):
#             fk_col = getattr(fk_target, self.fk_col)
#         else:
#             fk_col = Column(self.fk_col, pk_col.type, ForeignKey(pk_col))
#             setattr(fk_target, self.fk_col, fk_col)
#
#         rel = relationship(target_cls,
#                 primaryjoin=fk_col==pk_col,
#                 collection_class=self.kw.get('collection_class', set)
#             )
#         setattr(cls, key, rel)
#         self._setup_reverse(key, rel, target_cls)
#
# class one_to_many(FKRelationship):
#     """Generates a one to many relationship."""
#
#     def _get_pk_fk(self, cls, target_cls):
#         return cls, target_cls
#
# class many_to_one(FKRelationship):
#     """Generates a many to one relationship."""
#
#     def _get_pk_fk(self, cls, target_cls):
#         return target_cls, cls
#
# class many_to_many(DeferredProp):
#     """Generates a many to many relationship."""
#
#     def __init__(self, target, tablename, local, remote, **kw):
#         self.target = target
#         self.tablename = tablename
#         self.local = local
#         self.remote = remote
#         self.kw = kw
#
#     def _config(self, cls, key):
#         """Create an association table between parent/target
#         as well as a relationship()."""
#
#         target_cls = cls._decl_class_registry[self.target]
#         local_pk = list(cls.__table__.primary_key)[0]
#         target_pk = list(target_cls.__table__.primary_key)[0]
#
#         t = Table(
#                 self.tablename,
#                 cls.metadata,
#                 Column(self.local, ForeignKey(local_pk), primary_key=True),
#                 Column(self.remote, ForeignKey(target_pk), primary_key=True),
#                 keep_existing=True
#             )
#         rel = relationship(target_cls,
#                 secondary=t,
#                 collection_class=self.kw.get('collection_class', set)
#             )
#         setattr(cls, key, rel)
#         self._setup_reverse(key, rel, target_cls)
#
# # =====================================
# # end of zzzeek's code for "Magic" ORM
# # =====================================
class ASCII(TypeDecorator):
    '''
    A database string type that only allows ASCII characters.

    This is a data type check since all strings in the database could
    be unicode.
    '''
    impl = VARCHAR

    def process_bind_param(self, value, dialect):
        '''
        Run encode on a unicode string to make sure the string only
        contains ASCII characterset characters. To avoid another conversion
        pass, the original unicode value is passed to the underlying db.
        '''
        # run an encode on the value with ascii to see if it contains
        # non ascii. Ignore the return value because we only want to throw
        # an exception if the encode fails. The data can stay unicode for
        # insertion into the db.
        if value is not None:
            value.encode('ascii')
        return value

    def process_result_value(self, value, dialect):
        '''
        Run encode on a unicode string coming out of the database to turn the
        unicode string back into an encoded python bytestring.
        '''
        if isinstance(value, unicode):
            value = value.encode('ascii')
        return value


class JSONEncodedDict(TypeDecorator):
    '''
    Represents a dictionary as a json-encoded string in the database.
    '''
    impl = VARCHAR

    def process_bind_param(self, value, dialect):
        '''
        Turn a dictionary into a JSON encoded string on the way into
        the database.
        '''
        if value is not None:
            value = json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        '''
        Turn a JSON encoded string into a dictionary on the way out
        of the database.
        '''
        if value is not None:
            value = json.loads(value)
        return value


class ZipPickler(object):
    '''Simple wrapper for pickle that auto compresses/decompresses values'''
    def loads(self, string):
        return pickle.loads(zlib.decompress(string))

    def dumps(self, obj, protocol):
        return zlib.compress(pickle.dumps(obj, protocol))


class MutationDict(Mutable, dict):
    '''
    A dictionary that automatically emits change events for SQA
    change tracking.

    Lifted almost verbatim from the SQA docs.
    '''
    @classmethod
    def coerce(cls, key, value):
        "Convert plain dictionaries to MutationDict."
        if not isinstance(value, MutationDict):
            if isinstance(value, dict):
                return MutationDict(value)
            # this call will raise ValueError
            return Mutable.coerce(key, value)
        else:
            return value

    def update(self, *args, **kwargs):
        '''
        Updates the current dictionary with kargs or a passed in dict.
        Calls the internal setitem for the update method to maintain
        mutation tracking.
        '''
        for k, v in dict(*args, **kwargs).iteritems():
            self[k] = v

    def __setitem__(self, key, value):
        "Detect dictionary set events and emit change events."
        dict.__setitem__(self, key, value)
        self.changed()

    def __delitem__(self, key):
        "Detect dictionary del events and emit change events."
        dict.__delitem__(self, key)
        self.changed()

    def __getstate__(self):
        '''Get state returns a plain dictionary for pickling purposes.'''
        return dict(self)

    def __setstate__(self, state):
        '''
        Set state assumes a plain dictionary and then re-constitutes a
        Mutable dict.
        '''
        self.update(state)


class NotFound(NoResultFound):
    '''Generic Not Found Error'''
    pass


surrogate_pk_template = Column(Integer, nullable=False, primary_key=True)


class ModelMeta(sqlalchemy.ext.declarative.DeclarativeMeta):
    '''
    MetaClass that sets up some common behaviors for pybald models.

    Will assign tablename if not defined based on pluralization rules. Also
    applies project global table arguments. Lastly this will assign a
    numeric primary key id to the table if no primary key was defined
    in the model class.
    '''
    def __init__(cls, name, bases, ns):
        try:
            if Model not in bases:
                # Allow for inheritance of the Model object
                # for use with SqlAlchemy inheritance patterns
                super(ModelMeta, cls).__init__(name, bases, ns)
                return
        except NameError:
            # if this is the Model class, then just return
            return

        # set the tablename, if it's user set, use that, otherwise use a
        # function to create one
        cls.__tablename__ = getattr(cls, "__tablename__",
                                    pluralize(camel_to_underscore(name)))
        # tableargs adds autoload to create schema reflection
        cls.__table_args__ = getattr(cls, "__table_args__", {})

        if project.schema_reflection:
            # create or update the __table_args__ attribute
            cls.__table_args__['autoload'] = True
        if project.global_table_args:
            cls.__table_args__.update(project.global_table_args)

        # check if the class has at least one primary key
        # if not, automatically generate one.
        has_primary = any([False] + [value.primary_key
                                     for value in cls.__dict__.values()
                                     if isinstance(value, Column)])
        if not has_primary:
            # treat the id as a mixin w/copy
            surrogate_pk = surrogate_pk_template.copy()
            surrogate_pk._creation_order = surrogate_pk_template._creation_order
            cls.id = surrogate_pk
        super(ModelMeta, cls).__init__(name, bases, ns)


class RegistryMount(type):
    '''
    A registry creating metaclass that keeps track of all defined classes that
    inherit from a base class using this metaclass.
    '''
    # lifted almost verbatim from: http://martyalchin.com/2008/jan/10/simple-plugin-framework/
    def __init__(cls, name, bases, attrs):
        try:
            cls.registry.append(cls)
        except AttributeError:
            # this is processing the first class (the mount point)
            cls.registry = []

        return super(RegistryMount, cls).__init__(name, bases, attrs)


class RegistryModelMeta(RegistryMount, ModelMeta):
    pass


class Model(Base):
    '''Pybald Model class, inherits from SQLAlchemy Declarative Base.'''
    __metaclass__ = RegistryModelMeta

    NotFound = sqlalchemy.orm.exc.NoResultFound
    MultipleFound = sqlalchemy.orm.exc.MultipleResultsFound

    def is_modified(self):
        '''Check if SQLAlchemy believes this instance is modified.'''
        # TODO: Using this internal flag, is this a private internal
        # behavior? Is there a better, public-api way of getting this info?
        # alternatively is session.is_modified(self) although will this fail
        # if the instance isn't part of the session?
        return instance_state(self).modified

    def clear_modified(self):
        '''
        Unconditionally clear all modified state from the attibutes on
        this instance.
        '''
        # Uses this method: http://www.sqlalchemy.org/docs/orm/internals.html?highlight=commit_all#sqlalchemy.orm.state.InstanceState.commit_all
        if int(sa_min_ver) > 7:
            # the methods became private in SA 8, need to check if
            # this is a problem
            return instance_state(self)._commit_all({})
        else:
            return instance_state(self).commit_all({})

    def is_persisted(self):
        '''
        Check if this instance has ever been persisted. Means it is either
        `detached` or `persistent`.
        '''
        return has_identity(self)

    def save(self, flush=False):
        '''
        Saves (adds) this instance in the current databse session.

        This does not immediatly persist the data since these operations
        occur within a transaction and the sqlalchemy unit-of-work pattern.
        If something causes a rollback before the session is committed,
        these changes will be lost.

        When flush is `True`, flushes the data to the database immediately
        **within the same transaction**. This does not commit the transaction,
        for that use :py:meth:`~pybald.db.models.Model.commit`)
        '''

        session.add(self)

        if flush:
            self.flush()
        return self

    def delete(self, flush=False):
        '''
        Delete this instance from the current database session.

        The object will be deleted from the database at the next commit. If
        you want to immediately delete the object, you should follow this
        operation with a commit to emit the SQL to delete the item from
        the database and commit the transaction.
        '''
        session.delete(self)
        if flush:
            self.flush()
        return self

    def flush(self):
        '''
        Syncs all pending SQL changes (including other pending objects) to
        the underlying data store within the current transaction.

        Flush emits all the relevant SQL to the underlying store, but does
        **not** commit the current transaction or close the current
        database session.
        '''
        session.flush()
        return self

    def commit(self):
        '''
        Commits the entire database session (including other pending objects).

        This emits all relevant SQL to the databse, commits the current
        transaction, and closes the current session (and database connection)
        and returns it to the connection pool. Any data operations after this
        will pull a new database session from the connection pool.
        '''
        session.commit()
        return self

    @classmethod
    def get(cls, **where):
        '''
        A convenience method that constructs a load query with keyword
        arguments as the filter arguments and return a single instance.
        '''
        return cls.load(**where).one()

    @classmethod
    def all(cls, **where):
        '''
        Returns a collection of objects that can be filtered for
        specific collections.

        all() without arguments returns all the items of the model type.
        '''
        return cls.load(**where).all()

    # methods that return queries (must execute to retrieve)
    # =============================================================
    @classmethod
    def load(cls, **where):
        '''
        Convenience method to build a sqlalchemy query to return stored
        objects.

        Returns a query object. This query object must be executed to retrieve
        actual items from the database.
        '''
        if where:
            return session.query(cls).filter_by(**where)
        else:
            return session.query(cls)

    @classmethod
    def filter(cls, *pargs, **kargs):
        '''
        Convenience method that auto-builds the query and passes the filter
        to it.

        Returns a query object.
        '''
        return session.query(cls).filter(*pargs, **kargs)

    @classmethod
    def query(cls):
        '''
        Convenience method to return a query based on the current object
        class.
        '''
        return session.query(cls)

    @classmethod
    def show_create_table(cls):
        '''
        Uses the simple dump_engine to print to stdout the SQL for this
        table.
        '''
        cls.__table__.create(dump_engine)


class NonDbModel(object):
    '''
    A plain Python object that uses the RegistryMount system to register
    non database models. Inheriting from this allows the object to be registered
    as a Model.
    '''
    __metaclass__ = RegistryMount
    # use the same registry space for NonDbModels
    registry = Model.registry



#!/usr/bin/env python
# encoding: utf-8
# """
# Console.py
#
# Created by mikepk on 2011-10-16.
# Copyright (c) 2011 Michael Kowalchik. All rights reserved.
# """
import code

import readline
import atexit
import os

import webob
from ..test import Client


class Console(code.InteractiveConsole):
    '''Pybald console including history buffer per project.'''
    def __init__(self, project_name, package_name=None, app=None,
                 additional_symbols=None):
        if additional_symbols is None:
            additional_symbols = {}

        self.project_name = project_name
        if not package_name:
            package_name = self.project_name

        # add the application to the console namespace
        console_symbols = {'app': app}

        # add webob Req/Resp objects to the console for convenience
        console_symbols.update({"Request": webob.Request,
                                "Response": webob.Response})
        console_symbols.update({"c": Client(app=app)})
        console_symbols.update(additional_symbols)

        histfile = os.path.expanduser("~/.pybald-{0}-console-history".format(
                                                                 project_name))

        code.InteractiveConsole.__init__(self, locals=console_symbols,
                                                          filename="<console>")
        self.init_history(histfile)

    def init_history(self, histfile):
        readline.parse_and_bind("tab: complete")
        if hasattr(readline, "read_history_file"):
            try:
                readline.read_history_file(histfile)
            except IOError:
                pass
            atexit.register(self.save_history, histfile)

    def save_history(self, histfile):
        readline.write_history_file(histfile)

    def run(self):
        # Fire up the console with the project, controllers, and models defined.
        self.interact('''Welcome to the pybald interactive console\n'''
                      ''' ** project: {0} **'''.format(self.project_name))



#!/usr/bin/env python
# encoding: utf-8
from lxml import html

from webassets.env import Environment
from webassets import Bundle
from webassets.exceptions import BundleError
from pybald.core import page
import project
import os
from urlparse import urlparse

try:
    import pyhash
    hashfunc = pyhash.super_fast_hash()
except ImportError:
    hashfunc = hash

import logging
console = logging.getLogger(__name__)

# python 3
try:
    type(unicode)
except NameError:
    unicode = str

public_path = os.path.join(project.path or '', "public")

# set the bundle input and output paths
if project.BUNDLE_OUTPUT_PATH:
    bundle_output_path = project.BUNDLE_OUTPUT_PATH
else:
    bundle_output_path = ''

if project.BUNDLE_SOURCE_PATHS:
    bundle_input_paths = [os.path.join(project.path or '', path.lstrip('/'))
                            for path in project.BUNDLE_SOURCE_PATHS]
else:
    bundle_input_paths = [public_path]

cache_path = os.path.join(project.path or '', 'tmp', '.webassets-cache')
manifest_file = os.path.join(project.path or '', 'tmp', '.webassets-manifest')

# auto create cache path if not present
if not os.path.exists(cache_path):
    os.makedirs(cache_path)

# setting auto-build to false will keep all sub-nodes from
# running the XML parser.
env = Environment(public_path,
                  url='',
                  debug=(not project.BUNDLE_ASSETS),
                  auto_build=bool(project.BUNDLE_AUTO_BUILD),
                  load_path=bundle_input_paths,
                  cache=cache_path,
                  manifest="".join(["json:", manifest_file]),
                  # Take any bundle filter options and apply them to the config
                  **(project.BUNDLE_FILTER_OPTIONS or {}))

def _parse_bundle(elem, parent_bundle=None):
    '''Recursively generate webassets bundles by walking the xml/html tree'''
    if elem.tag != "bundle":
        raise SyntaxError("Bundling only works with bundle tags.")
    contents = []
    asset_type = None
    asset_types = []
    # get all sub_bundles and process them
    for t in elem:
        if t.tag == 'bundle':
            bundle, asset_type = _parse_bundle(t)
            contents.append(bundle)
        elif t.tag == 'link':
            file_path = urlparse(t.attrib.get('href', '')).path
            contents.append(file_path.lstrip('/'))
            asset_types.append('css')
        elif t.tag == 'script':
            file_path = urlparse(t.attrib.get('src', '')).path
            contents.append(file_path.lstrip('/'))
            asset_types.append('js')
    if not asset_types:
        raise SyntaxError("No assets found in the bundle tag. Include javascript, css, or source files in the bundle.")
    elif len(set(asset_types)) > 1:
        raise SyntaxError("All assets in a bundle must be of the same type. They should all be javascript, css etc... or compiled to the same asset type.")
    asset_type = asset_types[0]

    attrib_dict = dict(**elem.attrib)
    # convert text True/False to boolean
    if 'debug' in attrib_dict:
        if attrib_dict['debug'] == 'True':
            attrib_dict['debug'] = True
        elif attrib_dict['debug'] == 'False':
            attrib_dict['debug'] = False

    # default temporary output file, just the version as name...
    if 'output' not in attrib_dict:
        attrib_dict['output'] = '/min/{0}/%(version)s.{0}'.format(asset_type)

    # remove leading slashes since webassets works on relative paths
    # a little hacky but keeps the document (web page) syntax cleaner
    attrib_dict['output'] = os.path.join(bundle_output_path, attrib_dict['output'].lstrip('/')).lstrip('/')
    new_bundle = Bundle(*contents, **attrib_dict)
    return new_bundle, asset_type


class memoize_bundles(object):
    '''
    Store the result of the XML parse and return it on subsequent
    requests.
    '''
    def __init__(self, method):
        self.method = method
        self.cached = {}

    def __call__(self, input_text):
        if project.debug:
            return self.method(input_text)
        key = hashfunc(input_text)
        try:
            return self.cached[key]
        except KeyError:
            self.cached[key] = self.method(input_text)
            return self.cached[key]


@memoize_bundles
def bundle(input_text):
    '''
    Parse HTML/XML fragments from a string looking for asset bundles and
    process them into webasset bundles.
    '''
    # HTML/XML parse the fragments
    output_buffer = []
    html_fragment = u'''<div>{0}</div>'''.format(input_text)
    root = html.fragment_fromstring(html_fragment)

    for elem in root:
        # parse all bundles
        if elem.tag == 'bundle':
            bundle_tag = elem
            bundle, asset_type = _parse_bundle(bundle_tag)

            # every asset type has a default 'link type'
            link_func = getattr(page, 'add_{0}'.format(asset_type))
            # construct the asset urls
            # try:
            assets = [link_func(url) for url in bundle.urls(env=env)]
            output_buffer.extend(assets)
            # add any text nodes that got glommed onto
            # the node
            if bundle_tag.tail is not None:
                output_buffer.append(bundle_tag.tail)
        # just dump out anything that's not a bundle
        else:
            output_buffer.append(html.tostring(elem))
    return u'\n'.join([unicode(item) for item in output_buffer])



#!/usr/bin/env python
# encoding: utf-8

from webob import Response, exc
from pybald.core.templates import render as template_engine
import urllib
import logging
console = logging.getLogger(__name__)


class ErrorMiddleware:
    '''
    Handles exceptions during web transactions.

    Implemented as WSGI middleware, this middleware wraps the application and
    intercepts errors and exceptions. The middleware makes a distinction between
    webob HTTP exceptions and other exception types.


    :param application:  WSGI application/middleware that is to be
                         *wrapped* by the error handler in the web app pipeline.

    :param error_controller:  A class that is used by the middleware to handle
                          errors for the project. When configured, an error_controller
                          instance is created and the WSGI response is passed to
                          the error_controller to handle.
    '''
    def __init__(self, application=None, error_controller=None):
        if application:
            self.application = application
        else:
            # no pipeline so just generate a generic response
            self.applicaion = Response()
        self.error_controller = error_controller

    def __call__(self, environ, start_response):
        #pass through if no exceptions occur
        try:
            # catch stupid URLs before getting into webob
            try:
                unicode(environ['PATH_INFO'])
                urllib.unquote(environ['QUERY_STRING']).decode('utf8')
            except UnicodeDecodeError:
                return Response(status=400, body="""<h1>Bad Request</h1>""")(environ, start_response)
            return self.application(environ, start_response)
        # handle HTTP errors
        except exc.HTTPException, err:
            # if the middleware is configured with an error controller
            # use that to display the errors
            console.debug("HTTP Exception thrown {0}".format(err))
            if self.error_controller:
                handler = self.error_controller(status_code=err.code)

                try:
                    # try executing error_handler code
                    # otherwise re-raise the exception
                    return handler(environ, start_response)
                except Exception:
                    console.exception("Exception thrown during error_controller handling")
                    raise
            else:
                # HTTPExceptions are also WSGI apps and can be called as such
                return err(environ, start_response)
        except Exception, err:
            console.exception("General Exception thrown")
            if self.error_controller:
                handler = self.error_controller(message=str(err))
            else:
                # create a generic HTTP server Error webob exception
                handler = exc.HTTPServerError('General Fault')
            return handler(environ, start_response)


def pybald_error_template():
    '''
    This method is a shim between the old view error template function and the
    new template rendering methods. It should not be used and is present only
    to maintain backward compatibility.

    This is targeted for deprecation.
    '''
    console.warn("pybald_error_template is deprecated in pybald 0.2.0")
    return template_engine._get_template('stack_trace')



#!/usr/bin/env python
# encoding: utf-8

import os
import sys
import re

from inspect import isclass
from importlib import import_module

import logging
log = logging.getLogger(__name__)
import time


class PybaldImportError(ImportError):
    pass


PYTHON_MODULE_NAME_PATTERN = re.compile(r'^([a-z][0-9a-z_]*)\.py$', re.I)
PYTHON_MAGIC_VARIABLE_PATTERN = re.compile(r'^__.*__$')
import traceback


# TODO, use walk to have this recursively walk up the models path
# finding all interesting classes.
def load_all_from_path(path, package):
    '''
    '''
    for dirpath, dirnames, filenames in os.walk(path):
        if dirpath != path:
            # turn nested path into package notation
            nested = os.path.relpath(dirpath, path
                                        ).replace('/', '.').lstrip('.')
        else:
            nested = None
        for filename in sorted(filenames):
            match = PYTHON_MODULE_NAME_PATTERN.search(filename)
            if not match:
                continue
            import_module_name = match.group(1)
            try:
                if not nested:
                    import_package_name = import_module_name
                else:
                    import_package_name = '.'.join((nested, import_module_name))

                module = import_module(name="." + import_package_name,
                                       package=package.__name__)
                # print "import {1}{0}".format("." + import_package_name, package.__name__)
            except ImportError:
                if log.handlers or logging.getLogger().handlers:
                    log.exception("Automatic Pybald class loader failed with exception:")
                    log.error('-'*80)
                else:
                    tb = traceback.format_exc()
                    sys.stderr.write(tb + '-'*80 + '\n')

                raise PybaldImportError("\nThe automatic pybald class loader "
                                        "failed while attempting to load the module {0} from {1}.\n"
                                        "".format(filename, dirpath))


def flatten_namespace(package, registry):
    '''
    Export all classes in a registry into a package namespace.

    This is mainly a backwards compatibility function for exposing classes
    inside a flat package namespace (w/out additional module qualification).
    '''
    # log.debug("Flattening all registry classes into the package {0}...".format(package.__name__))
    for class_ in registry:
        setattr(package, class_.__name__, class_)
        package.__all__.append(class_.__name__)


def auto_load(package, registry=None):
    '''
    Walks the entire path of a package, recursively, and loads all python
    modules present within the package.

    Used for auto-loading classes and modules.
    '''
    package_path, filename = os.path.split(os.path.realpath(package.__file__))
    # if this is not a package, then return
    if not filename.startswith("__init__.py"):
        return
    t0 = time.time()
    load_all_from_path(package_path, package)
    if registry:
        flatten_namespace(package, registry)
    log.debug("Loaded all modules in {0:<30}- total load time: {1:0.2f}s".format(package.__name__, time.time() - t0))


def pybald_class_loader(path, classes, module_globals, module_locals,
                                                               recursive=True):
    '''
    Take a set of special class names, a path, and scan the path loading
    classes that match or inherit from the list. Then these are returned into
    the provided module scope.

    This is used by pybald to search for special "Controller" classes to load
    into the router object for the entire project. It's also used to auto-load
    all models in the project under the models namespace for less typing.

    This is a somewhat expensive function so is only executed on first startup
    of the project.

    Note: The second auto-model loading use of this function  may be too
    "magical" and may be deprecated in a future release.

    :param path: The path to scan for matching classes.

    :param classes: A list of classes to use for matching.

    :param module_globals: The globals member of the package from which this
         method is called.

    :param module_locals: The locals member of the package from which this
         method is called.

    :param recursve: Whether to scan the path recursively looking for
         matching classes.
    '''
    loaded_classes = []
    for dirpath, dirnames, filenames in os.walk(path):
        if dirpath != path:
            # turn nested path into package notation
            nested = os.path.relpath(dirpath, path
                                        ).replace('/', '.').lstrip('.')
        else:
            nested = None
        for filename in sorted(filenames):
            match = PYTHON_MODULE_NAME_PATTERN.search(filename)
            if not match:
                continue
            import_module_name = match.group(1)
            try:
                if not nested:
                    import_package_name = import_module_name
                    import_list = ()
                else:
                    import_list = (import_module_name,)
                    import_package_name = '.'.join((nested, import_module_name))
                scan_module = __import__(import_package_name,
                                         module_globals,
                                         module_locals,
                                         import_list,
                                         1)
            except ImportError:
                if log.handlers or logging.getLogger().handlers:
                    log.exception("Automatic Pybald class loader failed with exception:")
                    log.error('-'*80)
                else:
                    tb = traceback.format_exc()
                    sys.stderr.write(tb + '-'*80 + '\n')

                raise PybaldImportError("\nThe automatic pybald class loader "
                                        "failed while attempting to load the module {0} from {1}.\n"
                                        "".format(filename, dirpath))
            for classname in filter(lambda attribute_name: (
                    not PYTHON_MAGIC_VARIABLE_PATTERN.search(attribute_name) and
                                   attribute_name not in loaded_classes
                                   ),
                                                             dir(scan_module)):
                try:
                    module_class = getattr(scan_module, classname)
                    # only process class definitions from the modules
                    # not strictly necessary? Allow TypeError to catch non
                    # classes?
                    if not isclass(module_class):
                        continue
                    is_pybald_auto_loaded_module = issubclass(module_class,
                                                                       classes)
                # if the module member is not a class then skip
                except TypeError:
                    continue
                # if the module is one of the tracked class types
                # add the name to the provided scope
                if is_pybald_auto_loaded_module:
                    module_globals[classname] = module_class
                    loaded_classes.append(classname)
        if not recursive:
            break
    return loaded_classes



import project
import sys

def map(urls):
    '''
    Defines the url to controller mapping for the application. A 
    routes mapper or submapper object must be passed in.
    '''
    urls.connect('home', r'/', controller="home")

    # generic pattern
    urls.connect('base', r'/{controller}/{action}/{id}')
    urls.connect(r'/{controller}/{action}')
    urls.connect(r'/{controller}')

    # REDIRECT ALL URLS TERMINATING IN a slash, '/', to no slash
    # the opposite is also valid, all urls can be forced to have a slash
    # depends on your preference
    urls.redirect('{url:.*}/', '{url}',
                  _redirect_code='301 Moved Permanently')



#!/usr/bin/env python
# encoding: utf-8
"""
ErrorController.py

Created by mikepk on 2012-09-06.
Copyright (c) 2012 Michael Kowalchik. All rights reserved.
"""

import os
import unittest
from pybald.core.controllers import action, BaseController, render_view
from webob import Response
from mako import exceptions

import project
from collections import defaultdict


def error_response(template="general_error", status_code=500):
    template = os.path.join('error', template)
    return Response(body=render_view(template=template, format="html"),
                            status=status_code)


class ErrorController(BaseController):
    '''Controller to handle error exceptions.'''
    # map status codes to error controller actions
    error_map = defaultdict(lambda: 'general_error',
                                    {404: 'not_found',
                                     410: 'gone',
                                     401: 'not_authorized',
                                     500: 'general_error'})

    def __init__(self, *pargs, **kargs):
        '''Setup the error controller'''
        self.status_code = kargs.pop('status_code', 500)
        self.message = kargs.pop('message', None)
        super(ErrorController, self).__init__(*pargs, **kargs)

    @action
    def http_client_error(self, req):
        '''A normal, non-code, http error'''
        try:
            return error_response(self.error_map[self.status_code],
                                  self.status_code)
        except Exception, err:
            return error_response('general_error', self.status_code)

    @action
    def __call__(self, req):
        '''
        The ErrorController will return a formatted stack trace if in debug
        mode, or point to the regular error page otherwise.
        '''
        if project.email_errors or project.debug:
            stack_trace = render_view(template='stack_trace', data={'req': req})

        if project.debug:
            return Response(body=stack_trace, status=self.status_code)
        else:
            return error_response('general_error', self.status_code)


class ErrorControllerTests(unittest.TestCase):
    def setUp(self):
        pass

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# encoding: utf-8
from pybald.core.controllers import BaseController
from pybald.core import pybald_class_loader

__all__ = []


def load():
    '''Load all controllers inherting from BaseController'''
    global __all__
    controllers_path = __path__[0]
    __all__ = pybald_class_loader(controllers_path, (BaseController,), globals(), locals())



#!/usr/bin/env python
# encoding: utf-8
"""
SAGreen.py

Support code to allow SQLalchemy to operate correctly with Eventlet and Green Threads.

Created by mikepk on 2010-10-15.
Copyright (c) 2011 Michael Kowalchik. All rights reserved.
"""

import sqlalchemy.engine.url
from sqlalchemy.pool import QueuePool

from eventlet import timeout

import project
import eventlet

import os
os.environ["EVENTLET_THREADPOOL_SIZE"] = "20"
# importing tpool automatically spawns 20 native threads
# tunable with an env variable
from eventlet import tpool

class ConnectTimeout(Exception):
    '''Connection timeout.'''
    pass

def eventlet_greenthread_scope():
    '''Greenthread scope function for sqlalchemy scoped sessions.'''
    return id(eventlet.greenthread.getcurrent())

class GreenThreadQueuePool(QueuePool):
    '''Simple subclass of the standard SA QueuePool. Uses a sleep when the pool is empty to green thread switch.'''
    def do_get(self, *pargs, **kargs):
        # while the queue pool is filled, switch to another green thread until one is available
        # TODO: Add timeout code to make sure this doesn't spin forever and deadlock
        # if something goes wrong in the db connections
        # t = timeout.Timeout(20, ConnectTimeout())
        # try:
        # while self._max_overflow > -1 and self._overflow >= self._max_overflow
        # if wait:
        #     gt = eventlet.greenthread.getcurrent()
        #     gt.wait()
        # print "%s : sleeping/switching" % id(eventlet.greenthread.getcurrent())
        # print pargs, kargs
        # eventlet.sleep(0)
        # finally:
        #     # cancel the timeout
        #     t.cancel()
        return super(GreenThreadQueuePool,self).do_get(*pargs, **kargs)


# lifting some of the logic from sqlalchemy
# to reuse the dialect / url parsing to create
# connection arguments
def convert_url_to_connection_args(name_or_url, **kwargs):
    url = sqlalchemy.engine.url.make_url(name_or_url)
    dialect_cls = url.get_dialect()

    # get the correct DBAPI base on connection url
    dbapi_args = {}
    dbapi = dialect_cls.dbapi(**dbapi_args)

    # create the dialect
    dialect_args = {'dbapi':dbapi}
    dialect = dialect_cls(**dialect_args)

    # assemble connection arguments
    # for this dialect
    (cargs, connection_params) = dialect.create_connect_args(url)
    return dbapi, connection_params

def green_connection():
    '''Create a green databse connection.'''
    dbapi, kw = convert_url_to_connection_args(project.database_engine_uri)
    # TODO: this should check for supported dbapis before proceeding

    def eventlet_connection_creator():
        '''Creator method to wrap DBAPI connections with native threads to allow non-blocking db access.'''
        # This code is lifted and slightly tweaked
        # from the db_pool in the eventlet package
        t = timeout.Timeout(20, ConnectTimeout())
        try:
            conn = dbapi.connect(**kw)
            ## I think this is to keep the connection itself from blocking
            ## unfortunately it's hanging. The proxied connection still works
            ## nonblockingly though.
            # conn = tpool.execute(dbapi.connect, **kw)
            # conn.errorhandler = error_handle
            proxied_connection = tpool.Proxy(conn, autowrap_names=('cursor',))
            return proxied_connection
        finally:
            # cancel the timeout
            t.cancel()

    return eventlet_connection_creator




import re
import unicodedata
import sys
import os
import time
import signal
# first pass, anything before a CAPLower gets separated. i.e. CAP_Lower
#  123Lower -> 123_Lower
first_pass = re.compile(r'(.)([A-Z][a-z]+)')
# second pass, anything lowerCAP gets split then lowercased lower_cap
second_pass = re.compile(r'([a-z0-9])([A-Z])')


# technically this is PascalCase (or StudlyCaps)
def camel_to_underscore(text):
    '''Convert CamelCase text into underscore_separated text.'''
    return second_pass.sub(r'\1_\2', first_pass.sub(r'\1_\2', text)).lower()


def underscore_to_camel(text):
    '''Convert underscore_separated text into CamelCase text.'''
    return ''.join([token.capitalize() for token in text.split(r'_')])


ordinal_suffixes = {1: 'st', 2: 'nd', 3: 'rd'}


def ordinal_suffix(num):
    '''Returns the ordinal suffix of an interger (1st, 2nd, 3rd).'''
    if 10 <= abs(num) % 100 <= 20:
        return 'th'
    return ordinal_suffixes.get(abs(num) % 10, 'th')


plural_patterns = [(re.compile(pattern), re.compile(search), replace
                                        ) for pattern, search, replace in (
                         ('[^aeiouz]z$', '$', 's'),
                         ('[aeiou]z$', '$', 'zes'),
                         ('[sx]$', '$', 'es'),
                         ('[^aeioudgkprt]h$', '$', 'es'),
                         ('[^aeiou]y$', 'y$', 'ies'),
                         ('$', '$', 's'))]


def _build_plural_rule((pattern, search, replace)):
    return lambda word: pattern.search(word) and search.sub(replace, word)


plural_rules = map(_build_plural_rule, plural_patterns)


def pluralize(text):
    '''
    Returns a pluralized from of the input text following simple naive
    english language pluralization rules.

    Smart enough to turn fox into foxes and quiz into quizzes. Does not catch
    all pluralization rules, however.
    '''
    for rule in plural_rules:
        result = rule(text)
        if result:
            return result


def strip_accents(text):
    '''
    Strip diacriticals from characters. This will change the meaning of words
    but for places where unicode can't be used (or ASCII only) francais looks
    better than fran-ais or fran?ais.
    '''
    if isinstance(text, str):
        return unicode(text, errors='ignore')
    return ''.join((c for c in unicodedata.normalize('NFD', text) if
                                              unicodedata.category(c) != 'Mn'))


# ======
file_modification_times = {}


def is_modified(filename):
    '''
    Returns True if the file referenced by filename has been modified
    since the last time this funciton was called, otherwise False.
    '''
    new_modified_time = os.path.getmtime(filename)
    last_modified_time = file_modification_times.get(filename, None)
    if last_modified_time is None:
        file_modification_times[filename] = new_modified_time
        return False
    if last_modified_time < new_modified_time:
        file_modification_times[filename] = new_modified_time
        return True
    else:
        return False


def watch_module_files(pid=None):
    '''
    Continuously monitors all loaded modules and issues a POSIX SIGHUP if 
    any of the files associated with the loaded modules are changed during
    runtime.
    '''
    while True:
        changed = False
        for filename in filter(None, (getattr(module, '__file__', None) for
                                module in set(filter(None, sys.modules.values())))):
            changed = is_modified(filename)
            #print filename, changed
            if filename.endswith(".pyc") and os.path.exists(filename[:-1]):
                changed = changed or is_modified(filename[:-1])
                #print filename[:-1], changed
            if changed:
                print "File Changed!"

                os.kill(pid, signal.SIGHUP)
                # sys.exit()
        time.sleep(1)



#!/usr/bin/env python
# encoding: utf-8

#url_for takes these arguments as well
# anchor          specified the anchor name to be appened to the path
# host            overrides the default (current) host if provided
# protocol        overrides the default (current) protocol if provided
# qualified       creates the URL with the host/port information as
#                 needed

# TODO: add javascript escape code here so it's available in the template engine

from datetime import datetime
from routes import url_for
from mako import filters
from urlparse import urlparse, ParseResult
from routes import request_config
import project
try:
    import pyhash
    hashfunc = pyhash.super_fast_hash()
except ImportError:
    hashfunc = hash


# parse result keys
class AssetUrl(dict):
    '''
    Wraps urls and returns URL transformations when necessary.

    The primary use case is transforming the url for running static assets on a CDN.
    '''
    keys = ("scheme", "netloc", "path", "params", "query", "fragment")

    def __init__(self, url):
        self.raw_url = url
        super(AssetUrl, self).__init__(**dict(zip(self.keys, urlparse(url))))

    def __str__(self):
        return self.__html__()

    def __html__(self):
        '''Return a transformed URL if necessary (appending protocol and CDN)'''
        host = self.get('netloc', None)
        # Don't CDN urls with hosts we're not re-writing
        if host:
            if (project.STATIC_SOURCES is None or
                                           host not in project.STATIC_SOURCES):
                return self.raw_url
        if (project.USE_CDN and (project.CDN_HOST or project.STATIC_HOSTS)):
            # get the protocol for the current request
            # this requires the custom HTTP header X-Forwarded-Proto
            # set if running behind a proxy (or if SSL is terminated
            # upstream)
            protocol = request_config().protocol
            # use the round robin hosts to speed download when not https
            if protocol != "https" and project.STATIC_HOSTS:
                self['netloc'] = project.STATIC_HOSTS[hashfunc(self.raw_url) %
                                                      len(project.STATIC_HOSTS)]
            else:
                self['netloc'] = project.CDN_HOST
            # adjust the scheme of any link with a net location
            # to match the current request so we don't have mixed link
            # protocols
            if self['netloc']:
                self['scheme'] = protocol
        return ParseResult(**self).geturl()


class HTMLLiteral(unicode):
    '''This is an object to handle literal HTML in Mako template'''
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return self.__html__()

    def __html__(self):
        return self.value


def as_p(input_str):
    lines = input_str.splitlines()
    return unicode("".join([u"<p>{0}</p>".format(line) for line in lines]))


class tag(object):
    def set(self, **kargs):
        self.attribs.extend([u'''{0}="{1}"'''.format(k.rstrip('_'), v) for k, v in kargs.items()])
        return self


class img(tag):
    def __init__(self, src='', **kargs):
        self.img_src = src
        self.attribs = []
        self.set(**kargs)
        if "alt" not in kargs:
            self.set(alt=self.img_src)

    def __str__(self):
        return self.__html__()

    def __html__(self):
        '''Return the image in string form.'''
        return u'''<img src="{0}" {1} />'''.format(AssetUrl(str(self.img_src)),
                                                   " ".join(self.attribs))


class anchor(tag):
    def __init__(self, link_text='', name='', **kargs):
        self.link_text = filters.html_escape(link_text)
        if name:
            self.url = name
        else:
            self.url = self.link_text
        self.attribs = []
        self.set(**kargs)

    def __str__(self):
        return self.__html__()

    def __html__(self):
        '''Return the anchor in string form.'''
        # attr = " ".join(self.attribs)
        return u'''<a name="{0}" {1}>{2}</a>'''.format(self.url,
                                                       " ".join(self.attribs),
                                                       self.link_text)


class link(tag):
    def __init__(self, link_text='', **kargs):
        self.link_text = filters.html_escape(link_text)
        self.url = "#"
        self.attribs = []
        self.set(**kargs)

    def filter(self, filter_type="h"):
        self.link_text = filters.html_escape(self.link_text)
        return self

    def to(self, *pargs, **kargs):
        self.url = url_for(*pargs, **kargs)
        return self

    def to_raw_url(self, url):
        # Link directly with no url_for processing
        # (url_for can't deal with unicode!)
        self.url = url
        return self

    def __str__(self):
        return self.__html__()

    def __html__(self):
        '''Return the link in string form.'''
        attr = " ".join(self.attribs)
        return u'''<a href="{0}" {1}>{2}</a>'''.format(self.url, attr, self.link_text)


def plural(list_object):
    '''Return "s" for > 1 items'''
    if len(list_object) > 1:
        return "s"
    else:
        return ""


def humanize(date_string):
    format = "%Y-%m-%d %H:%M:%S"
    try:
        date = datetime.strptime(date_string, format)
    except:
        return date_string
    now = datetime.now()
    delta = now - date
    plural = 's'
    if delta.days < 0:
        return "in the future"
    elif delta.days >= 30:
        format = "%d %b"
        if date.year != datetime.now().year:
            format += " '%y"
        return str(date.strftime(format).lstrip('0'))
    elif delta.days >= 14:
        weeks = delta.days / 7
        if weeks == 1:
            plural = ''
        return "%s week%s ago" % (str(weeks), plural)
    elif delta.days >= 1:
        if delta.days == 1:
            plural = ''
        return "%s day%s ago" % (str(delta.days), plural)
    # > 1 hour, display in hours
    elif delta.seconds > 3600:
        hours = int(round(delta.seconds / 3600.0))
        if hours == 1:
            plural = ''
        return "%s hour%s ago" % (str(hours), plural)
    elif delta.seconds > 60:
        minutes = int(round(delta.seconds / 60.0))
        if minutes == 1:
            plural = ''
        return "%s minute%s ago" % (str(minutes), plural)
    else:
        return "just a moment ago"

# From django.utils.html: Javascript escape characters
_base_js_escapes = (
    ('\\', '\\u005C'),
    ('\'', '\\u0027'),
    ('"', '\\u0022'),
    ('>', '\\u003E'),
    ('<', '\\u003C'),
    ('&', '\\u0026'),
    ('=', '\\u003D'),
    ('-', '\\u002D'),
    (';', '\\u003B'),
    ('\u2028', '\\u2028'),
    ('\u2029', '\\u2029')
)

# From django.utils.html: Escape every ASCII character with a value less than 32.
_js_escapes = (_base_js_escapes +
               tuple([('%c' % z, '\\u%04X' % z) for z in range(32)]))


def js_escape(value):
    """
    Hex encodes characters for use in JavaScript strings.
    (from django.utils.html)
    """
    if value:
        for bad, good in _js_escapes:
            value = value.replace(bad, good)
    return value



#!/usr/bin/env python
# encoding: utf-8

import os

import project
from formalchemy import templates, config
from formalchemy import FieldSet as FAFieldSet, Field as FAField, Grid as FAGrid, validators

from pybald.core.templates import engine
from pybald.core.helpers import HTMLLiteral

import pybald.db.models
import inspect

# set the Pybald Mako engine to be the main
# form template engine
config.engine = engine.form_render


class Field(FAField):
    def render(self, *pargs, **kargs):
        output_data = super(Field, self).render(*pargs, **kargs)
        return HTMLLiteral(output_data)


class FieldSet(FAFieldSet):
    def render(self, *pargs, **kargs):
        output_data = super(FieldSet, self).render(*pargs, **kargs)
        return HTMLLiteral(output_data)


class BaseForm(FAFieldSet):
    def __init__(self, *pargs, **kargs):

        template = kargs.pop('template', 'fieldset')

        # This hack is to assign a session for non-instances. Used
        # primarily for loading the relations in models, otherwise
        # session has to be explicitly assigned (I think, can't find
        # docs to the contrary)
        if (pargs and isinstance(pargs[0], pybald.db.models.ModelMeta) and not
            'session' in kargs):
            kargs['session'] = pybald.db.models.session

        # Init the standard FieldSet
        super(BaseForm, self).__init__(*pargs, **kargs)
        # FieldSet.__init__(self,instance or self.__class__, data=data or None)
        # set the template_id to the name of the model
        self.template_id = os.path.join('forms', template)

    def render(self, *pargs, **kargs):
        output_data = super(BaseForm, self).render(*pargs, **kargs)
        return HTMLLiteral(output_data)


class MultiFieldSet(list):
    def __init__(self, *pargs, **kargs):
        super(MultiFieldSet, self).__init__()
        if pargs and inspect.isclass(pargs[0]):
            klass_name = pargs[0].__name__
        else:
            raise ValueError("SA Class required, instances not supported")

        self.fieldset_key = ''.join([klass_name, '--'])
        data = kargs.pop('data', None)
        if data is not None:
            keys = filter(lambda x: x.startswith(self.fieldset_key),
                                                                set(data.keys()))

            try:
                fields = [zip([key] * len(data.getall(key)), data.getall(key))
                                             for key in keys]
            except AttributeError:
                raise TypeError("data argument must support 'getall(key)' like WebOb MultiDict")

            transposed = zip(*fields)

            for data in [dict(x) for x in transposed]:
                self.append(BaseForm(data=data, *pargs, **kargs))
        else:
            self.append(BaseForm(data=None, *pargs, **kargs))





#!/usr/bin/env python
# encoding: utf-8

import os
import unittest
import project
from mako.template import Template
from mako.lookup import TemplateLookup
import re
import logging
console = logging.getLogger(__name__)

# base template helpers all pybald projects have
template_helpers = ['from pybald.core.helpers import img, link, humanize, js_escape, as_p',
                    'from pybald.core import page',
                    'from pybald.core.helpers import js_escape as js',
                    'from mako.filters import html_escape']

if project.template_helpers:
    template_helpers.extend(project.template_helpers)

# set the default filters to auto-html escape all content
default_filters = ['h', 'unicode']
if project.template_default_filters:
    default_filters = project.template_default_filters

# templates follow name.FORMAT.template so this is a simple
# regex check of that pattern
TEMPLATE_PATTERN = re.compile(r'([^\.]+)\.([^\.]+)\.template$')


class TemplateEngine:
    '''
    The basic template engine, looks up templates and renders them.
    Uses the mako template system
    '''

    def __init__(self, template_path=None, cache_path=None):
        self.project_path = project.path
        try:
            default_template_path = os.path.join(os.path.dirname(
                                                os.path.realpath(__file__)),
                                                'default_templates')
            project_template_path = template_path or os.path.join(
                                                            self.project_path,
                                                            'app/views')
            project_cache_path = cache_path or os.path.join(self.project_path,
                                                           'tmp/viewscache')
        except AttributeError:
            console.exception(("**Warning**\n"
                             "Unable to load templates from template path\n"
                             ))
            default_template_path = ""
            project_template_path = ""
            project_cache_path = ""

        fs_test = project.template_filesystem_check or project.debug or False

        self.lookup = TemplateLookup(directories=[project_template_path,
                                                  default_template_path],
                                     module_directory=project_cache_path,
                                     imports=template_helpers,
                                     input_encoding='utf-8',
                                     output_encoding='utf-8',
                                     filesystem_checks=fs_test,
                                     default_filters=default_filters)

    def form_render(self, template_name=None, format="form", **kargs):
        '''
        Render the form for a specific model using formalchemy.

        :param template_name: The name of the template to search for and
                              render for a form
        :param kargs: the data to render in the context of the form
                        template
        '''
        data = kargs
        try:
            template_id = kargs['fieldset'].template_id
        except (KeyError, AttributeError):
            template_id = 'forms/{0}'.format(template_name)

        mytemplate = self._get_template(template_id, format)
        # We use the render_unicode to return a native unicode
        # string for inclusion in another Mako template
        return mytemplate.render_unicode(**data)

    def _get_template(self, template, format="html"):
        '''
        Retrieves the proper template from the Mako template system.

        :param template: The name of the template in the filesystem to retrieve
                         for rendering.
        :param format: A string specifying the format for the template (html,
                       json, xml, etc...), overrides the format specified in
                       the data dictionary.

        The _get_template method of the template engine constructs a template
        name based on the template_id and the format and retrieves it from the
        Mako template system.
        '''
        # format can't be None, set to html as a default
        if format is None:
            format = 'html'
        if TEMPLATE_PATTERN.search(template):
            template_file = template
        else:
            template_file = "/{0}.{1}.template".format(template.lower().lstrip('/'),
                                                   format.lower())
        console.debug("Using template: {0}".format(template_file))
        # TODO: Add memc caching of rendered templates
        # also need to check if the internal caching is good enough
        return self.lookup.get_template(template_file)

    def __call__(self, template=None, data={}, format="html"):
        '''
        Renders the template.

        :param template: The name of the template in the filesystem to retrieve
                         for rendering.
        :param data: A dictionary that represents the context to render inside
                     the template. Keys in this dictionary will be available
                     to the template.
        :param format: A string specifying the format type to return (e.g. 'html',
                       'json', 'xml')

        Calls _get_template to retrieve the template and then renders it.
        '''
        template_data = dict(project.page_options.items() + data.items())
        mytemplate = self._get_template(template, format)
        console.debug("Rendering template")
        return mytemplate.render(**template_data)

# module scope singleton, should this be changed?
# engine = TemplateEngine()
render = TemplateEngine()


class CompatibilityProxy(object):
    '''A proxy to re-write the __call__ method.'''

    def __init__(self, obj):
        """The initializer."""
        self._obj = obj

    def _get_template(self, data, format="html", template=None):
        '''Old style get template'''
        template_name = template or data.pop('template_id', '')
        format = format or data.get('format', 'html')
        return self._obj._get_template(template=template_name, format=format)

    def __getattr__(self, attrib):
        return getattr(self._obj, attrib)

    def __call__(self, data, format=None, template=None):
        template_name = template or data.pop('template_id', '')
        format = format or data.get('format', 'html')
        return self._obj(template=template_name, data=data, format=format)

engine = CompatibilityProxy(render)


class TemplateEngineTests(unittest.TestCase):
    def setUp(self):
        pass

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# encoding: utf-8
'''HTML page helper functions as well as simple asset tag handling.'''

import os
import project

# from urlparse import urlparse, ParseResult
# global request_config... how can we eliminate?
# from routes import request_config
from pybald.core.helpers import HTMLLiteral, AssetUrl

import logging
console = logging.getLogger(__name__)
try:
    import pyhash
    hashfunc = pyhash.super_fast_hash()
except ImportError:
    console.warn("!"*10 + '''  Using python built-in hash() for asset URL
generation. This is system implementation specific and may result in different
hosts mapping static assets to different static hosts. That may cause
inefficient use of browser caches. Optionally you can install pyhash to
install additional fast, non-cryptographic, hashes that are not system
dependent.

pip install pyhash
''')
    hashfunc = hash



asset_tag_cache = {}


def compute_asset_tag(filename, pattern='{filename}{extension}?v={tag}'):
    asset_tag = asset_tag_cache.get(filename, None)
    try:
        if not asset_tag:
            asset_tag = str(
                int(round(os.path.getmtime(os.path.join(project.path,
                                                        "public",
                                                        filename.lstrip("/"))))
                ))
        asset_tag_cache[filename] = asset_tag
    except OSError:
        asset_tag_cache[filename] = "xxx"
    filename, ext = os.path.splitext(filename)
    return pattern.format(filename=filename, tag=asset_tag, extension=ext)


def add_js(filename):
    return HTMLLiteral('''<script src="{0}"></script>'''.format(
                                        AssetUrl(compute_asset_tag(filename))))


def add_css(filename, media="screen"):
    return HTMLLiteral('''<link type="text/css" href="{0}" media="{1}" rel="stylesheet">'''.format(
                                                AssetUrl(compute_asset_tag(filename)),
                                                str(media)))


def add_extern_css(filename, media="screen"):
    return HTMLLiteral(('''<link type="text/css" href="{0}"'''
            ''' media="{1}" rel="stylesheet">''').format(
                                                filename,
                                                str(media)))





#!/usr/bin/env python
# encoding: utf-8
import sys
import os
from pybald.core.class_loader import auto_load

# mod_wsgi restricts stdout, so switch it to stderr
original_stdout, sys.stdout = sys.stdout, sys.stderr

# find out where this module lives, set the project path accordingly
project_path, script_path = os.path.split(
                                os.path.dirname(os.path.realpath(__file__)))

# add the project and pybald paths to the sitedir
# TODO: make pybald installable so this is unnecesary
import site
site.addsitedir(project_path)
import project
site.addsitedir(project.toplevel)

# setup the package if not present, allows relative imports
if __name__ == '__main__' and __package__ is None:
    from importlib import import_module
    import_module('.wsgi', project.package_name)
    __package__ = project.package_name+'.wsgi'

from pybald.core.router import Router
from pybald.core.middleware.errors import ErrorMiddleware
# from pybald.core.middleware.sessions import SessionManager
# from pybald.core.middleware.users import UserManager
# from pybald.core.middleware.db_middleware import DbMiddleware

# load the application
# loading models and controllers
# auto_load'ing them ensures all models and controllers are loaded
# before proceeding
from ..app import urls, models, controllers
auto_load(models)
auto_load(controllers)

# load the error controller so that the ErrorMiddleware
# has somewhere to route the request if an error occurs
from ..app.controllers.error_controller import ErrorController
# from ..app.models.session import Session
# from ..app.models.user import User

# The main application pipeline
# Include all WSGI middleware here. The order of
# web transaction will flow from the bottom of this list
# to the top, and then back out. The pybald Router
# should usually be the first item listed.
# ----------------------------------
app = Router(controllers=controllers, routes=urls.map)
# app = UserManager(app, user_class=User)
# app = SessionManager(app, session_class=Session)
# app = DbMiddleware(app)
app = ErrorMiddleware(app, error_controller=ErrorController)
# ----------------------------------
#    ↑↑↑                  ↓↓↓
#    ↑↑↑                  ↓↓↓
#   Request              Response


# apache / mod_wsgi looks for 'application'
application = app


# called directly runs a simple dev server
# using the wsgiref server
def main():
    '''A simple server startup if module is called directly'''
    from optparse import OptionParser
    from wsgiref.simple_server import make_server
    from pybald.util.static_serve import StaticServer

    parser = OptionParser()
    parser.add_option("--host", default="0.0.0.0",
                      dest="host",
                      help="host ip to run")
    parser.add_option("-p", "--port",
                      type="int",
                      dest="port", default=8080,
                      help="the port to run on.")
    (options, args) = parser.parse_args()

    # add the static server component
    my_app = StaticServer(app, path=os.path.join(project_path, 'public'))
    httpd = make_server(options.host, options.port, my_app)
    print("Serving on {0}:{1}...".format(options.host, options.port))
    httpd.serve_forever()
    sys.exit(1)

#use when module called directly, runs internal webserver
if __name__ == '__main__':
    main()



from sqlalchemy.exc import SQLAlchemyError
from pybald.db import models
import sys


class EndPybaldMiddleware(object):
    '''Utilitiy middleware to force remove current session at the end of
    the request.'''
    def __init__(self, application):
        self.application = application

    def __call__(self, environ, start_response):
        try:
            return self.application(environ, start_response)
        finally:
            # always, always, ALWAYS close the session regardless
            models.session.remove()


class DbMiddleware(object):
    '''The database middleware provides three behaviors, committing
    transactions at the end of the request, rolling back database transactions
    if errors occur and forcibly closing the session
    at the end of the web request to avoid dangling connections.'''
    def __init__(self, application=None):
        if application:
            self.application = application
        else:
            # no pipeline so just generate a generic response
            self.applicaion = Response()

    def __call__(self, environ, start_response):
        # req = Request(environ)
        tb = None
        #pass through if no exceptions occur, commit sessions on complete
        try:
            resp = self.application(environ, start_response)
            # commit any outstanding sql
            models.session.commit()
        # on any SQLAlchemy Errors, rollback the transaction
        except SQLAlchemyError:
            # This pattern can cause memory leaks, so hopefully db errors
            # will be scrubbed from the code so this won't be hit
            excpt, detail, tb = sys.exc_info()
            models.session.rollback()

            # reraise the original details
            # can't use raw 'raise' because SA + eventlet
            # nukes sys_info
            raise excpt, detail, tb
        else:
            return resp
        finally:
            # always, always, ALWAYS close the session regardless
            # This remove() call has been moved higher in the WSGI stack
            # so that other things needing db sessions can still access the
            # db (like error reporting)
            # models.session.remove()
            del tb




"""
Automatically sets the environment variable `CELERY_LOADER` to
`pybald.celerey.loader:PybaldLoader`.  This ensures the loader is
specified when accessing the rest of this package, and allows celery
to be installed in a webapp just by importing pybald.celery::

    import pybald.celery

"""
import os
import warnings

CELERYPYBALD_LOADER = 'pybald.celery.loader.PybaldLoader'
if os.environ.get('CELERY_LOADER', CELERYPYBALD_LOADER) != CELERYPYBALD_LOADER:
    warnings.warn("'CELERY_LOADER' environment variable will be overridden by pybald.celery.")
os.environ['CELERY_LOADER'] = CELERYPYBALD_LOADER



#!/usr/bin/env python
# encoding: utf-8

import unittest
import re
from webob import Request, Response, exc
from routes import Mapper, request_config, URLGenerator
# handle Mako's top level lookup
from mako import exceptions
from pybald.util import camel_to_underscore
import logging
console = logging.getLogger(__name__)


class Router(object):
    # class method match patterns
    has_underscore = re.compile(r'^\_')
    controller_pattern = re.compile(r'(\w+)_controller')

    # add controllers=None to the call sig and use that for controller
    # loading
    def __init__(self, application=None, routes=None, controllers=None):
        '''
        Create a Router object, the core of the pybald framework.

        :param application:  WSGI application/middleware that is to be
                             *wrapped* by the router in the web app pipeline.

        :param routes: A routing function that takes a mapper (for parsing and
                       matching urls).

        :param controllers: A registry of all loaded controllers. This is
                            required to do routing lookups. This is also a
                            security precaution since only registered
                            controllers can be matched against.

        '''
        if routes is None or not callable(routes):
            raise TypeError("Route mapping is required. Please pass in a "
                            "routing function to the router as an arg to "
                            "Router init. "
                            "The routing function takes a routes mapper "
                            "object and uses it to contruct url mappings. "
                            "See pybald docs for more details.")

        self.controllers = {}
        # initialize Router
        # explicit turns off route memory and 'index' for
        # default action
        self.map = Mapper(explicit=False)
        routes(self.map)
        # debug print the whole URL map
        console.debug(str(self.map))
        if controllers is None:
            self.load_old_style()
        else:
            self.load(controllers)

    def load_old_style(self):
        '''Compatibility method for old-style pybald install.'''
        console.warn("!"*80+'''
!! Old style Router init is deprecated.
!!
!! New style pybald Router objects are constructed by passing in the controller
!! module or controller registry.
!! You will want to update your wsgi/myapp.py file.
!! app = Router(controllers=CONTROLLER_REGISTRY)
'''+"!"*80)
        import project
        # load the controllers from the project defined path
        # change this to passed in value to the Router. That way it can
        # be project specific
        # Load the project specified in the project file
        my_project = __import__(project.package_name, globals(), locals(),
                                                                    ['app'], -1)
        # add the project package name to the global symbol table
        # to avoid any double imports
        globals()[project.package_name] = my_project
        # import all controllers
        __import__('{project}.app'.format(project=project.package_name),
                              globals(), locals(), ['controllers'], -1)
        self.load(my_project.app.controllers)

    def load(self, controllers):
        '''
        Handles walking the registry (or old-style controller module) and
        builds the lookup table for controller classes to match against.

        Does some text munging to change the camel-case class names into
        underscore-separated url like names. (HomeController to home)

        :param controllers: A controller registry

        All controller candidates are loaded into a hash to look up
        the matched "controller" urlvar against.

        The _controller suffix is removed from the module name for the url
        route mapping table (so controller="home" matches home_controller).

        against and the routes regex is initialized with a list of controller
        names.

        Called only once at the start of a pybald application.
        '''

        controller_names = []
        # switch between old-style package container and registry model
        # both should work here
        if hasattr(controllers, '__all__'):
            controller_iterator = ((name, getattr(controllers, name)) for
                                   name in controllers.__all__)
        else:
            controller_iterator = ((controller.__name__, controller) for
                                   controller in controllers)
        for name, controller in controller_iterator:
            controller_name = camel_to_underscore(name)
            try:
                controller_path_name = self.controller_pattern.search(
                                                    controller_name).group(1)
            except AttributeError:
                controller_path_name = controller_name
            controller_names.append(controller_path_name)
            # self.controllers holds paths to map to modules and controller
            # names
            self.controllers[controller_path_name] = controller

        # register the controller module names
        # with the mapper, creates the internal regular
        # expressions
        self.map.create_regs(controller_names)

    def __repr__(self):
        return "<Pybald Router Object>"

    def get_handler(self, urlvars):
        '''Method that returns the callable code mapped to this current
        request.

        This method can be overriden to change the behavior of mapping.
        '''
        controller_name, action_name = urlvars["controller"], urlvars["action"]

        #methods starting with underscore can't be used as actions
        if self.has_underscore.match(action_name):
            raise exc.HTTPNotFound("Invalid Action")

        for key, value in urlvars.items():
            console.debug(u'''{0}: {1}'''.format(key, value))

        try:
            # create controller instance from controllers dictionary
            # using routes 'controller' returned from the match
            controller = self.controllers[controller_name]()
            handler = getattr(controller, action_name)
        # only catch the KeyError/AttributeError for the controller/action
        # search
        except (KeyError, AttributeError):
            raise exc.HTTPNotFound("Missing Controller or Action")

        return controller, handler

    def __call__(self, environ, start_response):
        '''
        A Router instance is a WSGI app. It accepts the standard WSGI call
        signature of ``environ``, ``start_response``.

        The Router has a few jobs. First it uses the Routes package to
        compare the requested path to available url patterns that have
        been loaded and passed to the Router upon init.

        Router is the most *framework-like* component of Pybald. In addition to
        dispatching urls to controllers, it also allows 'method override'
        behavior allowing other HTTP methods to be invoked such as ``put`` and
        ``delete`` from web clients that don't support them natively.

        :param environ: WSGI CGI-like request environment

        :param start_response: WSGI callback for starting the response and
                               setting HTTP response headers

        '''
        req = Request(environ)
        req.errors = 'ignore'
        #method override
        #===============
        # for REST architecture, this allows a POST parameter of _method
        # to be used to override POST with alternate HTTP verbs (PUT, DELETE)
        if req.POST:
            override_method = req.POST.pop('_method', None)
            if override_method is not None:
                environ['REQUEST_METHOD'] = override_method.upper()
                console.debug("Changing request method to {0}".format(
                                                        environ["REQUEST_METHOD"]))

        results = self.map.routematch(environ=environ)
        if results:
            match, route = results[0], results[1]
        else:
            match = route = None

        url = URLGenerator(self.map, environ)
        config = request_config()

        # Your mapper object
        config.mapper = self.map
        # The dict from m.match for this URL request
        config.mapper_dict = match
        config.host = req.host
        config.protocol = req.scheme
        # host_port
        # defines the redirect method. In this case it generates a
        # Webob Response object with the location and status headers
        # set
        config.redirect = lambda url: Response(location=url, status=302)

        environ.update({'wsgiorg.routing_args': ((url), match),
                        'routes.route': route,
                        'routes.url': url,
                        'pybald.router': self})

        # Add pybald extension
        # the pybald.extension is a dictionary that can be used to copy state
        # into a running controller (usually handled by the @action decorator)
        environ.setdefault('pybald.extension', {})["url_for"] = url

        # debug print messages
        console.debug('{0:=^79}'.format(' {0} '.format(req.path_qs)))
        console.debug('Method: {0}'.format(req.method))

        # use routes to match the url to a path
        # urlvars will contain controller + other non query string
        # URL data. Middleware above this can override and set urlvars
        # and the router will use those values.
        # TODO: allow individual variable overrides?
        urlvars = environ.get('urlvars', match) or {}

        # lifted from Routes middleware, handles 'redirect'
        # routes (map.redirect)
        if route and route.redirect:
            route_name = '_redirect_{0}'.format(id(route))
            location = url(route_name, **match)
            return Response(location=location,
                            status=route.redirect_status
                            )(environ, start_response)

        req.urlvars = urlvars
        environ['urlvars'] = urlvars

        if urlvars:
            controller, handler = self.get_handler(urlvars)
        # No URL vars means nothing matched in the mapper function
        else:
            raise exc.HTTPNotFound("No URL match")

        try:
            # call the action we determined from the mapper
            return handler(environ, start_response)
        # This is a mako 'missing template' exception
        except exceptions.TopLevelLookupException:
            raise exc.HTTPNotFound("Missing Template")
        # All other program errors are allowed to bubble up
        # e.g. a 500 server error


class routerTests(unittest.TestCase):
    def setUp(self):
        pass

    def testMap(self):
        router = Router()

    def testRoute(self):
        pass

    def testLoad(self):
        pass

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# encoding: utf-8
from pybald.db import models
from pybald.core import pybald_class_loader

__all__ = []


def load():
    '''Load all models and export them into this module's namespace'''
    global __all__
    models_path = __path__[0]
    __all__ = pybald_class_loader(models_path, (models.Model, models.NonDbModel), globals(), locals())


def create():
    """Create all models, issuing SQL to create the required storage representation."""
    models.Base.metadata.create_all()



from celery.loaders.base import BaseLoader
import site
from celery.utils import cwd_in_path
with cwd_in_path():
    import project
site.addsitedir(project.path)
site.addsitedir(project.toplevel)


def celery_config_item(item):
    # check if the config item starts with any of the celery config start keys
    # run startswith against each, and then "or" the results to return True
    # if any match
    # this could probably be done with a dynamic regex too.
    celery_config_start_keys = ('CELERY', 'BROKER')
    return reduce(lambda x, y: x or y,
                  map(item.startswith,
                      celery_config_start_keys))


class PybaldLoader(BaseLoader):
    """Pybald celery loader.

        * Maps the pybald.project config onto the celery config dict

    """
    def read_configuration(self):
        self.configured = True
        # generate a config dictionary from project.py
        return dict((key, getattr(project, key))
            for key in filter(celery_config_item, dir(project)))

    def on_worker_init(self):
        '''On worker init load all the default modules'''
        self.import_default_modules()



#!/usr/bin/env python
# encoding: utf-8
"""
project.py

A module to hold project specific members and configuration.
"""
import sys
import os
from pybald.core.logs import default_debug_log

# Configure the project path and package name
path = os.path.dirname(os.path.realpath(__file__))
toplevel = os.path.split(path)[0]
package_name = os.path.split(path)[-1]

# project name (otherwise default to the the project's path name)
project_name = package_name

# debug information
debug = True
env_name = "Development"

# options that are passed into the views
# directly
page_options = {}

# additional template helpers to add to the context for all templates
template_helpers = ['from pybald.core import page']

# route email to a local smtp server
smtp_config = {"smtp_server": "127.0.0.1",
               "smtp_port": 1025,
               "use_tls": False,
               "AuthUser": "USER@HOST",
               "AuthPass": ""}

# sqlalchemy engine string examples:
# mysql -         "mysql://{user}:{password}@{host}/{database}"
# postgres - postgresql://{username}:{password}@{host}:{port}/{database}'
# sqllite -       "sqlite:///{filename}"
# sqllite mem -   "sqlite:///:memory:"

# local database connection settings
# default to a sqllite file database based on the project name
database_engine_uri_format = 'sqlite:///{filename}'
db_config = {'filename': os.path.join(path,
             '{project}.sqlite'.format(project=project_name))}

# create the db engine uri
database_engine_uri = database_engine_uri_format.format(**db_config)

#sqlalchemy engine arguments
# database_engine_args = {'pool_recycle':3600,
#                         'pool_size':50,
#                         'max_overflow':9,
#                         'encoding':'utf-8' }
database_engine_args = {}

# Arguements applied to all SQLAlchemy tables
# useful mysql global args: {'mysql_engine':'InnoDB', 'mysql_charset':'utf8'}
global_table_args = {}

# check for the environment file, if there, override options
# with the environment
if os.path.isfile(os.path.join(path, "environment.py")):
    from environment import *
    sys.stderr.write("LOADED ENVIRONMENT: {0}\n".format(env_name))


# HACK: this allows project.X to return a default of None when
# undefined attributes are called (or setup from environment)
class ConfigWrapper(object):
    def __init__(self, wrapped):
        self.wrapped = wrapped

    def __getattr__(self, name):
        # Some sensible default?
        return getattr(self.wrapped, name, None)

    def __dir__(self):
        return dir(self.wrapped)

if debug:
    default_debug_log()

# Runs the project console. Allows interacting with the models and controllers.
if __name__ == '__main__':
    # import sys
    sys.path.append(toplevel)
    if __package__ is None:
        __import__(package_name, globals(), locals(), [], -1)
        __package__ = package_name

    # set the sys.modules project entry to this file
    # wrapped in the ConfigWrapper to avoid double import
    sys.modules['project'] = ConfigWrapper(sys.modules['__main__'])
    from pybald.util.console import Console

    # load the application!
    from .wsgi.myapp import app

    # now the models registry should be loaded and the additional_symbols
    # added
    from pybald.db.models import Model
    models = dict([(model.__name__, model) for model in Model.registry])

    # create a pybald console around it
    console = Console(project_name=project_name, app=app,
                      additional_symbols=models)
    console.run()
else:
    # Sets this module's entry in modules to be wrapped by the getattr
    # hack. Allows config options to be tested without an exception.
    sys.modules[__name__] = ConfigWrapper(sys.modules[__name__])



from sqlalchemy.interfaces import ConnectionProxy

# A simple way to wrap database transactions in either a query timer.
# or more interestingly an instrumented stack tracer that annotates
# queries with the calling function. Allows the MySQL slow query log
# to be uesed to find what code is issuing slow queries.

# import time
# import logging

# logging.basicConfig()
# logger = logging.getLogger("myapp.sqltime")
# logger.setLevel(logging.DEBUG)

class InstrumentedProxy(ConnectionProxy):
    def cursor_execute(self, execute, cursor, statement, parameters, context, executemany):
        try:
            import inspect
            stack = inspect.stack()
            # get the calling frame for the method/function calling
            # the sql command (taking into account the sql+execute functions)
            max_stack = len(stack) - 1
            print max_stack
            try:
                # for f in stack:
                #     frame = f[0]
                #     frame_info = inspect.getframeinfo(frame)
                #     debug = 'method:'+str(frame_info[2])+' file:'+str(frame_info[0])+' line:'+str(frame_info[1])
                #     print stack.index(f), debug

                frame = stack[14][0]
                frame_info = inspect.getframeinfo(frame)

                debug = 'method:'+str(frame_info[2])+' file:'+str(frame_info[0])+' line:'+str(frame_info[1])
                statement += ' -- '+debug
            except IndexError:
                pass
            return execute(cursor, statement, parameters, context)
        except:
            pass

# use by passing the proxy to the engine
#engine = sa.create_engine(project.database_engine_uri, proxy=InstrumentedProxy(), **engine_args)



from newrelic.agent import (current_transaction, wrap_function_wrapper,
                           FunctionTrace, WSGIApplicationWrapper)
from newrelic.api.function_trace import FunctionTraceWrapper
from functools import wraps, partial
import logging
console = logging.getLogger(__name__)

# define a new_relic_trace_wsgi function
nr_trace_wsgi = partial(FunctionTraceWrapper, group="Python/Pybald/WSGI")


class NewrelicPybald(object):
    '''
    Just a passthrough to remove the extra function definition from Newrelic
    when it instruments WSGI apps.

    This is a hack just for formatting purposes. It adds a layer to the WSGI
    stack which is a little non-ideal but avoids having to fork/hack newrelic's
    code.
    '''
    def __init__(self, application, name=None):
        if name:
            self.__class__.__name__ = name
        self.application = application

    def __call__(self, *args, **kargs):
        return self.application(*args, **kargs)


def handler_wrapper(name):
    def _inner_wrap(method):
        @wraps(method)
        def _wrapper(*args, **kargs):
            transaction = current_transaction()
            if transaction is None:
                return method(*args, **kargs)
            # name = "{0}:{1}".format(controller_name, action_name)
            # name = callable_name(wrapped)
            # console.debug(name)
            transaction.set_transaction_name(name)
            with FunctionTrace(transaction, name, group="Python"):
                return method(*args, **kargs)
        return _wrapper
    return _inner_wrap


def wrapper_Pybald_Router_get_handler(wrapped, instance, args, kwargs):
    '''
    Wrap the Pybald router's 'get handler' method with a function to
    automatically set the transaction name to the pybald contoller/action
    id.

    Monkeypatches the Router and allows us to dynamically wrap the return
    handler action on return.
    '''
    controller, action = wrapped(*args, **kwargs)
    if controller:
        controller_name = controller.__class__.__name__
    else:
        controller_name = ""
    action_name = action.__name__
    name = "{0}:{1}".format(controller_name, action_name)
    action = handler_wrapper(name)(action)
    return controller, action


def add_wsgi_trace(app):
    '''Recursively walk the WSGI stack adding new relic trace wrappers.

    :param app: A WSGI application to apply newrelic instrumentation to
    '''
    try:
        next_level = app.application
        if next_level is None:
            return True
        else:
            app.application = nr_trace_wsgi(next_level)
    except AttributeError:
        return True
    return add_wsgi_trace(next_level)


def instrument_pybald(app, name=None):
    '''
    Add the WSGI Transaction wrapper for newrelic.

    Also recursively walks
    up the WSGI stack and adds FunctionTraceWrappers to all of the WSGI apps
    above the instrumentation point.

    :param app: A WSGI application to apply newrelic instrumentation to
    :param name: The (optional) name of the application
    '''
    import pybald
    app = NewrelicPybald(app, name)
    add_wsgi_trace(app)
    app = WSGIApplicationWrapper(app, name=name, group="Python",
                                 framework=("Pybald", pybald.__version__))
    return app


def instrument_pybald_app(app, name=None):
    '''
    Monkeypatch the Pybald Router and instrument the WSGI stack with
    FunctionTrace wrappers.

    This allows the breakdown of all of the components in the Pybald stack and
    also sets the transaction names to the controller/action pairs for
    user code.

    :param app: A WSGI application to apply newrelic instrumentation to
    :param name: The (optional) name of the application
    '''
    import pybald.core.router
    wrap_function_wrapper(pybald.core.router, 'Router.get_handler',
                          wrapper_Pybald_Router_get_handler)
    return instrument_pybald(app, name=name)



import sqlalchemy as sa
import project
import sys

engine_args = project.database_engine_args

if project.green:
    from SAGreen import green_connection, GreenThreadQueuePool
    engine_args["creator"] = green_connection()
    engine_args["poolclass"] = GreenThreadQueuePool


def dump(sql, *multiparams, **params):
    print str(sql.compile(dialect=dump_engine.dialect) )

dump_engine = sa.create_engine('mysql://', strategy='mock', executor=dump)

try:
    engine = sa.create_engine(project.database_engine_uri, **engine_args)
except AttributeError:
    sys.stderr.write("**WARNING**\nSQLALchemy/pybald is using a mock"
                     " db connection.\n")
    engine = dump_engine

